#include <string.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <complex>
#include <cmath>

using namespace std;
typedef complex<double> Complex;

void semi_inf(int, Complex, Complex*, Complex* , Complex* );
void gleft(int,Complex,Complex*,Complex*,Complex*,Complex*);

extern "C"{
	void zgetrf_(const long &m, const long &n, Complex *a, const long &lda,
			long *ipiv, long &info);
	void zgetri_(const long &n, Complex *a, const long &lda,long *ipiv,
			Complex *work, const long &lwork, long &info);
	void zgemm_(const char &transa, const char &transb,
			const long &m, const long &n, const long &k,
			const Complex &alpha, Complex *a, const long &lda,
			Complex *b, const long &ldb, const Complex &beta,
			Complex *c, const long &ldc);
	void zgeev_(const char &jobvl, const char &jobvr, const long &n, Complex *A, const long &lda, Complex* w, Complex* vl, const long &ldvl, Complex* vr, const long &ldvr, Complex* work, const long &lwork, double* rwork, long &info);
}


/* 
 dim = dim(epsilon) = dim(t)
 epsilon: on-site potential
 t: hopping
 g: output Green's function
*/
void semi_inf(int dim, Complex z, Complex* epsilon, Complex* t, Complex* g){
	Complex *X, *ti;
	X  = new Complex[4*dim*dim];
	ti = new Complex[dim*dim]; // inverse matrix of t
	
	memcpy (ti, t, sizeof(Complex)*dim*dim );

	// inverse matrix of t
	// ti = t^{-1}
	long *ipiv, info=0, lwork=dim;
	ipiv = new long[dim];
	Complex *work;
	work = new Complex[lwork];	
	zgetrf_(dim,dim,ti,dim,ipiv,info);
	if(info){
		cout << "zgetrf error" << info << endl;
		exit(1);
	}
	zgetri_(dim,ti,dim,ipiv,work,lwork,info);
	if(info){
		cout << "zgetri error" << info << endl;
		exit(1);
	}
	if((double)lwork/(double)real(work[0])<1){
		cout << "zgetri error  lwork/maxlwork = " << lwork/real(work[0]) << endl;
		exit(1);
	}
	
	// g0i = g^{-1} = z-epsilon
	Complex *g0i;
	g0i = new Complex[dim*dim];
	for(int i=0; i<dim*dim; i++){
		g0i[i] = -epsilon[i];  
	}
	for(int i=0; i<dim; i++){
		g0i[i+i*dim] += z;
	}
	
	// tmpC = g^{-1} t^{-1}
	Complex *tmpC;
	tmpC = new Complex[dim*dim];
	zgemm_('N','N',dim,dim,dim,
			1,g0i,dim,ti,dim,0,tmpC,dim);

	//// start X-matrix
	// t^{-1} -> X
	for(int i=0; i<dim; i++){
		for(int j=0; j<dim; j++){
			X[dim*(2*dim)+i+j*(2*dim)] = ti[i+j*dim]; 
		}
	}
	// -t^\dag -> X
	for(int i=0; i<dim; i++){
		for(int j=0; j<dim; j++){
			X[dim+i+j*(2*dim)] = -conj(t[j+i*dim]); 
		}
	}
	// g^{-1} t^{-1} (=tmpC) -> X
	for(int i=0; i<dim; i++){
		for(int j=0; j<dim; j++){
			X[dim+dim*(2*dim)+i+j*2*dim] = tmpC[i+j*dim];
		}
	}
	//// end X-matrix
	
	//diagonalization of X
	long dimX = 2*dim; 
	lwork=4*dimX;
	double *rwork;
	rwork = new double[2*dimX];
	Complex *evalX,*evecX,*work2;
	evalX = new Complex[dimX];
	evecX = new Complex[dimX*dimX];
	work2 = new Complex[lwork];
	zgeev_('N','V',dimX,X,dimX,evalX,evecX,dimX,evecX,dimX,work2,lwork,rwork,info);
	if(info){
		cout << "zgeev error" << endl;
		exit(1);
	}
	if(real(work2[0])/lwork > 1.0){
		cout << "zgeev error" << endl;
		exit(1);
	}
	
	// sort (bubble sort)
	// evalX (evecX) is (not) overwritten
	Complex tmp, *pevecX[dimX], *tmpp;
	for(int i=0; i<dimX; i++){
		pevecX[i] = &evecX[dimX*i];
	}
	for(int j=dimX-1; j>0; j--){
	for(int i=0; i<j; i++){
		if(abs(evalX[i])>abs(evalX[i+1])){
			tmp = evalX[i+1];
			evalX[i+1] = evalX[i];
			evalX[i] = tmp;
			tmpp = pevecX[i+1];
			pevecX[i+1] =pevecX[i];
			pevecX[i] = tmpp;
		}
	}
	}
		
	//Matrices U12 and U22
	Complex *U12, *U22;
	U12 = new Complex[dim*dim];
	U22 = new Complex[dim*dim];
	for(int j=0; j<dim; j++){
		for(int i=0; i<dim; i++){
			U12[i+j*dim] = *(pevecX[j+dim]+i);
			U22[i+j*dim] = *(pevecX[j+dim]+i+dim);
		}
	}
	
	// U22 <- U22^{-1}
	zgetrf_(dim,dim,U22,dim,ipiv,info);
	if(info){
		cout << "zgetrf error" << info << endl;
		exit(1);
	}
	zgetri_(dim,U22,dim,ipiv,work,lwork,info);
	if(info){
		cout << "zgetri error" << info << endl;
		exit(1);
	}
	if((double)lwork/(double)real(work[0])<1){
		cout << "zgetri error  lwork/maxlwork = " << lwork/real(work[0]) << endl;
		exit(1);
	}
	
	// g = U12 U22^{-1}
	zgemm_('N','N',dim,dim,dim,1,U12,dim,U22,dim,0,g,dim);
	
	// free
	delete [] X,g0i,ti,work,tmpC,rwork,evalX,evecX,work2,U12,U22,pevecX;
}

/* 
 dim = dim(epsilon) = dim(t)
 epsilon: on-site potential at the (l-1)-th site
 t: hopping from the l-th site to the (l-1)-th site
 g0: input:  the left Green's function at the (l-1)-th site
 g1: output: the left Green's function at the     l-th site
*/
void gleft(int dim, Complex z, Complex* epsilon, Complex* t, Complex* g0, Complex* g1){
	Complex *tmp1;
	tmp1 = new Complex[dim*dim];
	
	// tmp1 = (z-epsilon)
	for(int i=0; i<dim*dim; i++){
		tmp1[i] = - epsilon[i];
	}
	for(int i=0; i<dim; i++){
		tmp1[i+i*dim] += z;
	}
		
	Complex *tmp2;
	tmp2 = new Complex[dim*dim];
	// tmp2 = -t^\dag g0
	zgemm_('C','N',dim,dim,dim,-1.0,t,dim,g0,dim,0,tmp2,dim);
	// tmp1 = tmp1 -t^\dag g0 t
	zgemm_('N','N',dim,dim,dim,1.0,tmp2,dim,t,dim,1.0,tmp1,dim);
	
	// tmp1 = tmp^{-1}
	long *ipiv, info=0, lwork=2*dim;
	ipiv = new long[dim];
	Complex *work;
	work = new Complex[lwork];	
	zgetrf_(dim,dim,tmp1,dim,ipiv,info);
	if(info){
		cout << "zgetrf error" << info << endl;
		exit(1);
	}
	zgetri_(dim,tmp1,dim,ipiv,work,lwork,info);
	if(info){
		cout << "zgetri error" << info << endl;
		exit(1);
	}
	if((double)lwork/(double)real(work[0])<1){
		cout << "zgetri error  lwork/maxlwork = " << lwork/real(work[0]) << endl;
		exit(1);
	}
	
	// g1 = tmp1
	memcpy(g1,tmp1,sizeof(Complex)*dim*dim);
	
	delete [] tmp1,tmp2,ipiv,work;
}